# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini
"""This module includes dataclasses to for parameters used in analysis of experiments."""

from dataclasses import dataclass
from typing import Optional

from abvelocity.param.derived_expt_stats import DerivedExptStats

UNION_MERGE = "union"
"""This merging method will concatenate the dataframes of the experiments vertically.
This is useful if an experiment is run on different segments and the user wants to analyze them together.
This will simply take the union of the assignment data vertically.
TODO: This is not implemented yet.
"""
CROSS_MERGE = "cross"
"""This merging method will create a cross variant for each unit of experiment.
For example if

    - experiment 1 has variants A and B
    - experiment 2 has variants C and D

Then the cross merge will create 9 variants because we do allow for "nan"s in the assignment data (to denote no trigger for that experiment).
The set of all possible cross-variants are:

    - (A, C)
    - (A, D)
    - (A, nan)
    - (B, C)
    - (B, D)
    - (B, nan)
    - (nan, C)
    - (nan, D)
    - (nan, nan)

"""

IMPLEMENTED_MERGE_METHODS = (CROSS_MERGE,)


@dataclass
class ExptInfo:
    """
    This is a dataclass to contain the information for one experiment in the analysis.
    """

    name: Optional[str] = None
    """
    name of the experiment. This is used merely in reports etc.
    If None, test_key will be used in it's place.
    """
    test_key: Optional[str] = None
    """
    Test key of the experiment.
    """
    experiment_id: Optional[int] = None
    """
    Experiment (iteration) ID.
    """
    segment_id: Optional[int] = None
    """
    Segment ID.
    """
    hash_id: Optional[int] = None
    """
    hash ID for the experiment (hashes units to variants).
    This field should be optional in most use cases.
    """
    start_date: Optional[str] = None
    """
    Start Date in this format: YYYY-MM-dd-00"""
    end_date: Optional[str] = None
    """
    End date in this format: YYYY-MM-dd-00.
    """
    variant_map: Optional[dict[str, str]] = None
    """
    A dictionary to map a variant name to a new name.
    This is useful to combine various variants into one,
    or improving on existing naming in data.
    As an example an unusual control label can be converetd here.
    """
    variants: Optional[tuple[str]] = None
    """
    The variants that the user is interested in.
    This is an optional field and can be utilized
    to only choose needed variants.
    """
    control_label: Optional[str] = "control"
    """
    The key for the control arm (variant).
    That key will be mapped to "control" in the SQL query.
    """
    treatment_label: Optional[str] = "enabled"
    """
    The key for the treatment arm (variant).
    That key will be mapped to "enabled" in the SQL query.
    # TODO: This field is not used in the current implementation.
    """
    query: Optional[str] = None
    """
    The query to get experiment assignment data.
    If this field is specified, then this will be used to get the data.
    This is especially useful to specify more complex queries to get assignmenmt data.
    For example if one experiment data is stored across test keys / segments etc. That might require a complex join and this field allows users to
    pass their own query.
    """
    derived_stats: Optional[DerivedExptStats] = None
    """
    Derived statistics for the experiment.
    This field can only be populated after the expt assignment data is extracted.
    """

    def __str__(self):
        variants_str = ", ".join(self.variants) if self.variants else "None"
        derived_stats_str = str(self.derived_stats) if self.derived_stats else "None"
        return (
            f"Experiment Information:\n"
            f"- Test Key: {self.test_key}\n"
            f"- Experiment ID: {self.experiment_id}\n"
            f"- Segment ID: {self.segment_id}\n"
            f"- Hash ID: {self.hash_id}\n"
            f"- Start Date: {self.start_date}\n"
            f"- End Date: {self.end_date}\n"
            f"- Variants Mapping: {self.variant_map}\n"
            f"- Variants: {variants_str}\n"
            f"- Control Label: {self.control_label}\n"
            f"- Treatment Label: {self.treatment_label}\n"
            f"- query: {self.query}\n"
            f"- Derived Stats:\n{derived_stats_str}\n"
        )

    def to_html(self):
        variants_str = ", ".join(self.variants) if self.variants else "None"
        derived_stats_html = self.derived_stats.to_html() if self.derived_stats else "<p>None</p>"
        return (
            f"<div>"
            f"<h3>Experiment Information</h3>"
            f"<p><strong>Test Key:</strong> {self.test_key}</p>"
            f"<p><strong>Experiment ID:</strong> {self.experiment_id}</p>"
            f"<p><strong>Segment ID:</strong> {self.segment_id}</p>"
            f"<p><strong>Hash ID:</strong> {self.hash_id}</p>"
            f"<p><strong>Start Date:</strong> {self.start_date}</p>"
            f"<p><strong>End Date:</strong> {self.end_date}</p>"
            f"<p><strong>Variants Mapping:</strong> {self.variant_map}</p>"
            f"<p><strong>Variants:</strong> {variants_str}</p>"
            f"<p><strong>Control Label:</strong> {self.control_label}</p>"
            f"<p><strong>Treatment Label:</strong> {self.treatment_label}</p>"
            f"<p><strong>Query:</strong> {self.query}</p>"
            f"<h4>Derived Stats:</h4>{derived_stats_html}"
            f"</div>"
        )

    def to_markdown(self) -> str:
        """
        Convert the dataclass fields and values to a Markdown format with headings and values.
        """
        variants_str = ", ".join(self.variants) if self.variants else "None"
        derived_stats_md = self.derived_stats.to_markdown() if self.derived_stats else "None"

        markdown_str = "### Experiment Information\n"
        markdown_str += f"- **Test Key**: {self.test_key}\n"
        markdown_str += f"- **Experiment ID**: {self.experiment_id}\n"
        markdown_str += f"- **Segment ID**: {self.segment_id}\n"
        markdown_str += f"- **Hash ID**: {self.hash_id}\n"
        markdown_str += f"- **Start Date**: {self.start_date}\n"
        markdown_str += f"- **End Date**: {self.end_date}\n"
        markdown_str += f"- **Variants Mapping**: {self.variant_map}\n"
        markdown_str += f"- **Variants**: {variants_str}\n"
        markdown_str += f"- **Control Label**: {self.control_label}\n"
        markdown_str += f"- **Treatment Label**: {self.treatment_label}\n"
        markdown_str += f"- **Query**: {self.query}\n"
        markdown_str += f"### Derived Stats\n{derived_stats_md}\n"

        return markdown_str

    def __post_init__(self):
        if self.name is None:
            if self.test_key is not None:
                self.name = self.test_key
            else:
                self.name = "Expt"


@dataclass
class MultiExptInfo:
    expt_info_list: list[ExptInfo]
    """A list of single experiments info."""
    merge_method: Optional[str] = None
    """The method to merge the individual experiments data.
    The only supported method is `CROSS_MERGE = "cross"` which will create tuples
    for each unit to represent the exposed variants (or no trigger) for each of
    experiments given in `expt_info_list`.
    """
    derived_stats: Optional[DerivedExptStats] = None
    """
    The derived statistics for the multi-experiment.
    """

    def __post_init__(self):
        """Checks to see if merge method is implemented."""
        if self.merge_method is not None and self.merge_method not in IMPLEMENTED_MERGE_METHODS:
            raise NotImplementedError(f"Merge method {self.merge_method} is not implemented.")

    def __str__(self):
        expt_info_str = "\n\n".join(str(expt) for expt in self.expt_info_list)
        derived_stats_str = str(self.derived_stats) if self.derived_stats else "None"
        return (
            f"Multiple Experiment Information:\n"
            f"{expt_info_str}\n"
            f"- Merge Method: {self.merge_method}\n"
            f"- Derived Stats:\n{derived_stats_str}"
        )

    def to_html(self):
        """
        Convert the MultiExptInfo data to a html format.
        """
        expt_info_html = "".join(expt.to_html() for expt in self.expt_info_list)
        derived_stats_html = self.derived_stats.to_html() if self.derived_stats else "<p>None</p>"
        return (
            f"<div>"
            f"<h3>Multiple Experiment Information</h3>"
            f"{expt_info_html}"
            f"<p><strong>Merge Method:</strong> {self.merge_method}</p>"
            f"<h4>Derived Stats:</h4>{derived_stats_html}"
            f"</div>"
        )

    def to_markdown(self) -> str:
        """
        Convert the MultiExptInfo data to a Markdown format.
        """
        expt_info_md = "\n".join(expt.to_markdown() for expt in self.expt_info_list)
        derived_stats_md = self.derived_stats.to_markdown() if self.derived_stats else "None"

        markdown_str = (
            f"### Multiple Experiment Information\n"
            f"{expt_info_md}\n"
            f"- **Merge Method**: {self.merge_method}\n"
            f"### Derived Stats\n{derived_stats_md}\n"
        )

        return markdown_str



from typing import Dict, List, Optional, Union

import pandas as pd


def prep_df_for_grouped_plot(
    df: pd.DataFrame, x_col: str, y_col: str, group_by_cols: Optional[List[str]] = None
) -> Dict[str, Union[pd.DataFrame, List[str]]]:
    """
    Transforms a DataFrame to prepare it for plotting multiple lines/markers
    based on unique combinations of grouping columns.

    If `group_by_cols` are provided, the DataFrame is pivoted such that each
    unique combination of grouping columns becomes a new column, with `x_col`
    as the index. The values for these new columns are taken from `y_col`.

    If `group_by_cols` are not provided, the original DataFrame is returned
    with only `x_col` and `y_col`, and `y_col` retains its original name.

    Args:
        df (pd.DataFrame): The input DataFrame.
        x_col (str): The name of the column to be used as the x-axis.
        y_col (str): The name of the column to be used as the y-axis (values).
        group_by_cols (Optional[List[str]]): A list of column names to group by.
            Each unique combination of these
            columns will result in a separate
            line/marker series. If None, no
            grouping/pivoting is performed.

    Returns:
        Dict[str, Union[pd.DataFrame, List[str]]]: A dictionary containing:
            - 'df' (pd.DataFrame): The transformed DataFrame suitable for `plot_lines_markers`.
                The column names for the new 'y' series will be descriptive
                of the group combinations.
            - 'y_cols' (List[str]): A list of the names of the new 'y' columns that were created
                or identified for plotting (e.g., `[original_y_col_name]` if no grouping).

    Raises:
        ValueError: If `x_col` or `y_col` are not found in the DataFrame.
        ValueError: If any column in `group_by_cols` is not found in the DataFrame.
    """

    # Validate essential columns
    if x_col not in df.columns:
        raise ValueError(f"x_col '{x_col}' not found in DataFrame columns.")
    if y_col not in df.columns:
        raise ValueError(f"y_col '{y_col}' not found in DataFrame columns.")

    if group_by_cols:
        # Validate group_by_cols
        for col in group_by_cols:
            if col not in df.columns:
                raise ValueError(f"Group-by column '{col}' not found in DataFrame columns.")

        # Ensure all columns needed for pivoting are present
        cols_for_pivot = group_by_cols + [x_col, y_col]
        filtered_and_sorted_df = df[cols_for_pivot]

        # Handle potential non-unique index/columns for pivoting
        # Sort by x_col and group_by_cols to ensure consistent ordering
        sort_cols = [x_col] + group_by_cols
        filtered_and_sorted_df = filtered_and_sorted_df.sort_values(by=sort_cols).reset_index(
            drop=True
        )

        # Convert group_by_cols to categorical to ensure all categories are present as columns
        # even if they have no data for certain x_col values after pivoting.
        for col in group_by_cols:
            # Filter out NaN values from categories before creating Categorical
            all_categories = pd.Series(df[col].unique()).dropna().tolist()
            filtered_and_sorted_df[col] = pd.Categorical(
                filtered_and_sorted_df[col], categories=all_categories
            )

        # Pivot the DataFrame
        pivot_df = filtered_and_sorted_df.pivot_table(
            index=x_col,
            columns=group_by_cols,
            values=y_col,
            dropna=False,  # Keep columns even if they become entirely NaN after pivot
            observed=False,  # Silence FutureWarning, ensure all categories are observed
        )

        # Flatten multi-level column index if group_by_cols has multiple columns
        if isinstance(pivot_df.columns, pd.MultiIndex):
            new_y_cols = ["_".join(map(str, col)).strip() for col in pivot_df.columns.values]
        else:
            # For a single group_by_col, the column name is just the value
            new_y_cols = [str(col).strip() for col in pivot_df.columns.values]

        pivot_df.columns = new_y_cols

        # Reset index to make x_col a regular column again
        pivot_df = pivot_df.reset_index()

        return {"df": pivot_df, "y_cols": new_y_cols}
    else:
        # No grouping, just select x_col and y_col, keeping original y_col name
        result_df = df[[x_col, y_col]]
        new_y_cols = [y_col]  # Keep original y_col name
        return {"df": result_df, "y_cols": new_y_cols}



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini


import os
from typing import Any, Dict, List, Optional

from abvelocity.get_data.cursor import Cursor


def desc_table(
    cursor: Cursor,
    table_name: str,
    count_distinct_cols: Optional[List[str]] = None,
    distinct_values_cols: Optional[List[str]] = None,
    log_file: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Describes a table by retrieving metadata, number of rows, and optionally calculates
    count distinct and distinct values for specified columns.

    Args:
        cursor: A database cursor (already connected) with `execute` method.
        table_name (str): The name of the table to describe.
        count_distinct_cols (Optional[List[str]]): List of column names to calculate COUNT(DISTINCT).
        distinct_values_cols (Optional[List[str]]): List of column names to retrieve distinct values.
        log_file (Optional[str]): File path to store the printed results. Default is None.

    Returns:
        Dict: A dictionary containing table metadata and optional calculations.
    """
    desc_dict = {"table_name": table_name}

    try:
        # Retrieve column metadata and row count
        cursor.execute(f"DESCRIBE {table_name}")
        columns_info = cursor.fetchall()

        desc_dict["columns"] = [{"name": col[0], "type": col[1]} for col in columns_info]

        # Fetch the total number of rows in the table
        cursor.execute(f"SELECT COUNT(*) FROM {table_name}")
        total_rows = cursor.fetchone()[0]
        desc_dict["number_of_rows"] = total_rows

        # Fetch the first 10 rows of the table
        cursor.execute(f"SELECT * FROM {table_name} LIMIT 10")
        desc_dict["rows"] = cursor.fetchall()

        # Calculate COUNT(DISTINCT) for specified columns
        if count_distinct_cols:
            count_distinct = {}
            for column in count_distinct_cols:
                cursor.execute(f"SELECT COUNT(DISTINCT {column}) FROM {table_name}")
                distinct_count = cursor.fetchone()[0]
                count_distinct[column] = distinct_count
            desc_dict["count_distinct"] = count_distinct

        # Retrieve distinct values for specified columns
        if distinct_values_cols:
            distinct_values = {}
            for column in distinct_values_cols:
                cursor.execute(f"SELECT DISTINCT {column} FROM {table_name}")
                values = [row[0] for row in cursor.fetchall()]
                distinct_values[column] = values
            desc_dict["distinct_values"] = distinct_values

    except Exception as e:
        desc_dict["error"] = str(e)

    # Convert results to string format
    result_str = "\n".join([f"{key}: {value}" for key, value in desc_dict.items()])

    # Print results
    print(result_str)

    # Write results to log file if provided
    if log_file:
        os.makedirs(os.path.dirname(log_file), exist_ok=True)
        with open(log_file, "w") as f:
            f.write(result_str + "\n")

    return desc_dict



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini

from typing import Dict, List, Optional, Union

import pandas as pd
import plotly.graph_objects as go

# Assuming these are imported from their respective modules
from abvelocity.utils.color_utils import get_distinct_colors
from abvelocity.utils.prep_df_for_grouped_plot import prep_df_for_grouped_plot


def plot_lines_markers(
    df: pd.DataFrame,
    x_col: str,
    line_cols: Optional[List[str]] = None,
    marker_cols: Optional[List[str]] = None,
    band_cols: Optional[
        List[str]
    ] = None,  # Corrected: Expects list of column names where each column contains tuples
    band_cols_dict: Optional[
        Dict[str, List[str]]
    ] = None,  # Expects dict where value is [lower_bound_col, upper_bound_col]
    line_colors: Optional[List[str]] = None,
    marker_colors: Optional[List[str]] = None,
    band_colors: Optional[List[str]] = None,
    title: Optional[str] = None,
) -> go.Figure:
    """Creates a Plotly figure with lines, markers, and/or filled bands.

    This function provides a lightweight and easy-to-use interface to visualize
    data from a DataFrame, allowing for the plotting of multiple series with
    customizable colors and legends.

    Args:
        df (pd.DataFrame): DataFrame containing the data to plot. It must include
            `x_col` and any columns specified in `line_cols`, `marker_cols`,
            `band_cols`, or `band_cols_dict`.
        x_col (str): The name of the column to use for the x-axis.
        line_cols (Optional[List[str]]): A list of column names whose values
            will be plotted as lines/curves on the y-axis. Defaults to None.
        marker_cols (Optional[List[str]]): A list of column names whose values
            will be plotted as markers/points on the y-axis. Defaults to None.
        band_cols (Optional[List[str]]): A list of column names. Each column
            specified here is expected to contain tuples, where each tuple
            represents the (lower bound, upper bound) of a band at a given
            x-value. Defaults to None.
        band_cols_dict (Optional[Dict[str, List[str]]]): A dictionary where
            keys are the desired legend names for bands, and values are lists
            containing two column names representing the lower and upper bounds
            of a band (e.g., `{"forecast": ["forecast_lower", "forecast_upper"]}`).
            Defaults to None.
        line_colors (Optional[List[str]]): A list of color strings to use for
            `line_cols`. If provided, its length must be at least the number
            of `line_cols`. Defaults to None.
        marker_colors (Optional[List[str]]): A list of color strings to use for
            `marker_cols`. If provided, its length must be at least the number
            of `marker_cols`. Defaults to None.
        band_colors (Optional[List[str]]): A list of color strings to use for
            `band_cols` or `band_cols_dict`. These colors will be used as the
            fill color for each band. If provided, its length must be at least
            the number of bands. Defaults to None, in which case distinct colors
            are generated with 0.2 opacity.
        title (Optional[str]): The title of the plot. Defaults to None,
            resulting in no title.

    Returns:
        go.Figure: An interactive Plotly graph figure.

    Raises:
        ValueError: If `line_colors`, `marker_colors`, or `band_colors` lists
            are provided but are shorter than the corresponding column lists.
        ValueError: If none of `line_cols`, `marker_cols`, `band_cols`, or
            `band_cols_dict` are provided.
    """

    if line_colors is not None and line_cols is not None:
        if len(line_colors) < len(line_cols):
            raise ValueError(
                "If `line_colors` is passed, its length must be at least `len(line_cols)`"
            )

    if marker_colors is not None and marker_cols is not None:
        if len(marker_colors) < len(marker_cols):
            raise ValueError(
                "If `marker_colors` is passed, its length must be at least `len(marker_cols)`"
            )

    if band_colors is not None and band_cols is not None:
        if len(band_colors) < len(band_cols):
            raise ValueError(
                "If `band_colors` is passed, its length must be at least `len(band_cols)`"
            )

    if band_colors is not None and band_cols_dict is not None:
        if len(band_colors) < len(band_cols_dict):
            raise ValueError(
                "If `band_colors` is passed, its length must be at least `len(band_cols_dict)`"
            )

    if line_cols is None and marker_cols is None and band_cols is None and band_cols_dict is None:
        raise ValueError(
            "At least one of `line_cols` or `marker_cols` or `band_cols`"
            " or `band_cols_dict` must be passed as a list (not None)."
        )

    fig = go.Figure()
    # Below we count the number of figure components to assign proper labels to legends.
    count_fig_data = -1
    if line_cols is not None:
        for i, col in enumerate(line_cols):
            line_properties = (
                go.scatter.Line(color=line_colors[i])
                if line_colors is not None
                else go.scatter.Line()
            )

            fig.add_trace(
                go.Scatter(
                    x=df[x_col], y=df[col], mode="lines", line=line_properties, showlegend=True
                )
            )
            count_fig_data += 1
            fig["data"][count_fig_data]["name"] = col

    if marker_cols is not None:
        for i, col in enumerate(marker_cols):
            marker_properties = (
                go.scatter.Marker(color=marker_colors[i])
                if marker_colors is not None
                else go.scatter.Marker()
            )
            fig.add_trace(
                go.Scatter(
                    x=df[x_col],
                    y=df[col],
                    mode="markers",
                    marker=marker_properties,
                    showlegend=True,
                )
            )
            count_fig_data += 1
            fig["data"][count_fig_data]["name"] = col

    if band_cols is not None:
        if band_colors is None:
            band_colors = get_distinct_colors(num_colors=len(band_cols), opacity=0.2)

        for i, col_name in enumerate(band_cols):  # Changed 'col_tuple' to 'col_name'
            # Ensure line properties are defined for band outlines, even if transparent
            line_properties_band = go.scatter.Line(color="rgba(0, 0, 0, 0)")

            fig.add_traces(
                [
                    go.Scatter(
                        x=df[x_col],
                        y=df[col_name].apply(lambda b: b[1]),  # Corrected to apply on column
                        mode="lines",
                        line=line_properties_band,
                        showlegend=False,  # Don't show legend for the invisible upper line
                    ),
                    go.Scatter(
                        x=df[x_col],
                        y=df[col_name].apply(lambda b: b[0]),  # Corrected to apply on column
                        mode="lines",
                        line=line_properties_band,
                        fill="tonexty",
                        fillcolor=band_colors[i],
                        showlegend=True,
                    ),
                ]
            )

            # Assign legend name to the filled band trace
            # The last added trace (index len(fig["data"]) - 1) is the filled lower bound trace
            fig["data"][len(fig["data"]) - 1]["name"] = col_name  # Use the column name as legend

    if band_cols_dict is not None:
        if band_colors is None:
            band_colors = get_distinct_colors(num_colors=len(band_cols_dict), opacity=0.2)

        for i, name in enumerate(band_cols_dict):
            col1 = band_cols_dict[name][0]
            col2 = band_cols_dict[name][1]
            # Ensure line properties are defined for band outlines, even if transparent
            line_properties_band = go.scatter.Line(color="rgba(0, 0, 0, 0)")

            fig.add_traces(
                [
                    go.Scatter(
                        x=df[x_col],
                        y=df[col2],
                        mode="lines",
                        line=line_properties_band,
                        showlegend=False,  # Don't show legend for the invisible upper line
                    ),
                    go.Scatter(
                        x=df[x_col],
                        y=df[col1],
                        mode="lines",
                        line=line_properties_band,
                        fill="tonexty",
                        fillcolor=band_colors[i],
                        showlegend=True,
                    ),
                ]
            )
            # Assign legend name to the filled band trace
            # The last added trace (index len(fig["data"]) - 1) is the filled lower bound trace
            fig["data"][len(fig["data"]) - 1]["name"] = name

    fig.update_layout(title=title)
    return fig


def plot_long_df(
    df: pd.DataFrame,
    x_col: str,
    y_col: str,
    group_by_cols: Optional[List[str]] = None,
    line_colors: Optional[List[str]] = None,
    title: Optional[str] = None,
) -> Dict[str, Union[go.Figure, pd.DataFrame, List[str]]]:
    """Plots data from a "long" DataFrame, transforming it to a "wide" format.

    This function prepares your data using `prep_df_for_grouped_plot` and then
    generates a Plotly figure where `y_col` values are plotted against `x_col`.
    If `group_by_cols` are provided, a separate line is drawn for each unique
    combination of these grouping columns.

    Args:
        df (pd.DataFrame): The input DataFrame in a "long" format.
        x_col (str): The name of the column to be used as the x-axis.
        y_col (str): The name of the column to be used as the y-axis (values).
        group_by_cols (Optional[List[str]]): A list of column names to group by.
            Each unique combination of these columns will result in a separate line.
            If None, a single line representing `y_col` vs `x_col` is plotted.
        line_colors (Optional[List[str]]): A list of color strings to use for each
            line. If None, `plot_lines_markers` will use default Plotly colors.
        title (Optional[str]): The plot title.

    Returns:
        Dict[str, Union[go.Figure, pd.DataFrame, List[str]]]: A dictionary containing:
            'fig' (go.Figure): The generated Plotly interactive figure.
            'df' (pd.DataFrame): The transformed DataFrame (in "wide" format)
                used for plotting.
            'y_cols' (List[str]): A list of the names of the columns that were
                plotted as lines.
    """

    prepared_data = prep_df_for_grouped_plot(
        df=df, x_col=x_col, y_col=y_col, group_by_cols=group_by_cols
    )
    wide_df = prepared_data["df"]
    y_cols = prepared_data["y_cols"]

    # Pass the wide_df to plot_lines_markers, as it contains the pivoted data
    fig = plot_lines_markers(
        df=wide_df,
        x_col=x_col,
        line_cols=y_cols,  # These are the new columns created by prep_df_for_grouped_plot
        line_colors=line_colors,
        title=title,
    )

    return {"fig": fig, "df": wide_df, "y_cols": y_cols}



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini
"""The goal is to simulate experiment data."""

import copy
from typing import Optional, TextIO

import numpy as np
import pandas as pd

from abvelocity.get_data.data_container import DataContainer
from abvelocity.get_data.join_expt_dfs import join_expt_dfs
from abvelocity.param.constants import CATEG_NAN_VALUE, UNIT_COL, VARIANT_COL


class Sim:
    """
    The goal is to simulate experiment data.

        - it allows to introduce attributes e.g. country etc in the population
        - it allows for overlapping multiple experiments
        - it allows to add baseline metrics which depend on the attributes (but not the experiment assignments)
        - it allows for adding univariate experiment impacts
        - it allows for adding interaction effects (multi-experiment)

    # Here is an example simulation

    # Population attributes
    attribute_weights = {
        "Level": {"Senior": 0.5, "Junior": 0.5},
        "Country": {"Iran": 0.1, "Canada": 0.3, "UK": 0.2, "US": 0.4},
        "Device": {"Smartphone": 0.6, "Tablet": 0.3, "Laptop": 0.1},
    }

    # Example parameters for multiple experiments
    expt_variant_weights_multi = [
        {"control": 0.4, "v1": 0.3, "v2": 0.3},  # Experiment 1
        {"control": 0.3, "enabled": 0.7},  # Experiment 2
    ]

    # Baseline metrics which depends on attributes (no expt is applied yet)
    metric_attribute_values = {
        "metric1": {"Level": {"Junior": 10, "Senior": 5}},
        "metric2": {"Country": {"USA": 20, "Canada": 15}},
    }

    # Univariate Effects
    # For metric1: v1 is bad; enabled is bad. v2 is good.
    # For metric2: v2 is good. control is good.
    expt_metric_impacts = [
        {
            "control": {"metric1": 1.5},
            "v1": {"metric1": -2, "metric2": +5},
            "v2": {"metric1": +5},
        },  # Experiment 1 impact
        {"control": {"metric2": 1}, "enabled": {"metric1": -2, "metric2": -5}},  # Experiment 2 impact
    ]

    # Interactions
    # Only metric1 has interactions.
    # The interactions will push up impact of (v1, enabled)
    # and we expect it to change the results due to univar effect for metric1

    interaction_metric_impacts = {("v1", "enabled"): {"metric1": 15}, ("v2", "control"): {"metric1": -2}}

    sim = Sim(
        population_size=1000,
        attribute_weights=attribute_weights,
        metric_attribute_values=metric_attribute_values,
        expt_variant_weights_multi=expt_variant_weights_multi,
        population_pcnt_multi=[90, 90],
        non_trigger_pct_multi=[5, 5],
        population_seed=42,
        expt_assignment_seed_multi=[13, 17],
        expt_metric_impacts=expt_metric_impacts,
        interaction_metric_impacts=interaction_metric_impacts,
    )

    sim.run()

    assert len(sim.population_df) == 1000

    for expt_data in sim.expt_data_list:
        expt_df = expt_data[0]
        assert len(expt_df) == 900

    assert sim.expt_df.shape == (985, 7)
    assert sim.expt_metric_df.shape == (985, 9)

    # Get a summary of how the data looks like.
    sim.publish(file=None)

    """

    def __init__(
        self,
        population_size: int = 0,
        attribute_weights: dict[str, dict[str, float]] = dict(),
        metric_attribute_values: dict[str, dict[str, dict[str, float]]] = dict(),
        expt_variant_weights_multi: tuple[dict[str, float]] = tuple(),
        population_pcnt_multi: tuple[float] = tuple(),
        non_trigger_pct_multi: tuple[float] = tuple(),
        population_seed: Optional[int] = None,
        expt_assignment_seed_multi: tuple[int] = tuple(),
        expt_metric_impacts: tuple[dict[str, float]] = tuple(),
        interaction_metric_impacts: dict[tuple[str], dict[str, float]] = dict(),
    ):
        # Based on input
        self.population_size = population_size
        self.attribute_weights = attribute_weights
        self.metric_attribute_values = metric_attribute_values
        self.expt_variant_weights_multi = expt_variant_weights_multi
        self.population_pcnt_multi = population_pcnt_multi
        self.non_trigger_pct_multi = non_trigger_pct_multi
        self.population_seed = population_seed
        self.expt_assignment_seed_multi = expt_assignment_seed_multi
        self.expt_metric_impacts = expt_metric_impacts
        self.interaction_metric_impacts = interaction_metric_impacts

        # Attributes generated by the `run` method
        # Population dataframe
        self.population_df = None
        self.population_summary = None

        # A list of tuples with first element being the expt dataframe
        # second element: summary of that data
        self.expt_data_list = None

        # The join experiment assignment data (no metrics)
        self.expt_df = None
        self.expt_df_summary = None

        # Baseline metric data
        self.expt_metric_baseline_df = None

        # Joined data of experiments and
        self.expt_metric_df = None

    def generate_population(
        self,
        population_size: int,
        attribute_weights: dict[str, dict[str, float]],
        seed: Optional[int] = None,
    ) -> tuple[pd.DataFrame, dict[str, dict[str, float]]]:
        """
        Generates a population DataFrame based on attribute weights and computes the attribute proportions.

        Args:
            population_size: The size of the population to generate.
            attribute_weights: A dictionary where keys are attribute names and values
                are dictionaries of attribute values with their corresponding probabilities.
            seed: An optional seed for the random number generator. Defaults to None.

        Returns:
            result: A tuple containing the population DataFrame and
                a dictionary with attribute proportions.
        """
        rng = np.random.default_rng(seed)

        data = {UNIT_COL: list(range(1, population_size + 1))}

        # Generate distributions for each attribute based on weights
        for attribute, weights in attribute_weights.items():
            choices = list(weights.keys())
            probs = list(weights.values())
            distribution = rng.choice(choices, p=probs, size=population_size)
            data[attribute] = distribution

        # Create DataFrame
        population_df = pd.DataFrame(data)

        # Compute summaries
        summary = {}
        for attribute in attribute_weights.keys():
            attribute_summary = population_df[attribute].value_counts(normalize=True).to_dict()
            summary[f"{attribute} Proportion"] = attribute_summary

        return population_df, summary

    def add_baseline_metrics(
        self, df: pd.DataFrame, metric_attribute_values: dict[str, dict[str, dict[str, float]]]
    ) -> pd.DataFrame:
        """
        Adds baseline metrics to a DataFrame based on attribute values and coefficients.

        Args:
            df: A DataFrame containing the population data.
            metric_attribute_values:
                A nested dictionary where the first-level keys are metric names, the second-level
                keys are attribute names, and the third-level keys are attribute values with their
                corresponding coefficients. Defaults to None.

        Returns:
            result: The original DataFrame with added baseline metrics.
        """
        if metric_attribute_values:
            for metric_name, coeffs in metric_attribute_values.items():
                metric_values = np.zeros(len(df))
                for attribute, attribute_coeffs in coeffs.items():
                    for value, coefficient in attribute_coeffs.items():
                        metric_values += np.where(df[attribute] == value, coefficient, 0)
                        df[metric_name] = metric_values

        return df

    def simulate_expt_assignment(
        self,
        population_df: pd.DataFrame,
        expt_variant_weights: dict[str, float],
        population_pcnt: float = 100.0,
        non_trigger_pct: float = 0.0,
        seed: Optional[int] = None,
    ) -> tuple[pd.DataFrame, dict[str, dict[str, float]]]:
        """
        Simulates experiment assignment for a sample of the population.

        Args:
            population_df: A DataFrame containing the population data.
            expt_variant_weights: A dictionary where keys are experiment variants and values are their
                corresponding probabilities.
            population_pcnt: The percentage of the population to be sampled for the experiment.
            non_trigger_pct: The percentage of the sample that should not be assigned to any experiment
                variant (denoted by a specific value). Defaults to 0.0.
            seed: An optional seed for the random number generator. Defaults to None.

        Returns:
            result: A tuple containing the sampled DataFrame with experiment
                assignments and a dictionary summarizing the proportions of each variant.
        """
        rng = np.random.default_rng(seed)

        sample_size = int(len(population_df) * (population_pcnt / 100.0))
        expt_df = population_df.sample(n=sample_size, random_state=seed)

        # Perform deep copy to preserve the original input
        expt_variant_weights = copy.deepcopy(expt_variant_weights)

        if non_trigger_pct > 0.0:
            expt_variant_weights[CATEG_NAN_VALUE] = non_trigger_pct / 100.0
            # We need to make sure the weights still sum up to 1
            # To correct that we need to correct the weights for all variants except for `CATEG_NAN_VALUE`
            # We can only allocate (100 - non_trigger_pct) percent to them
            # So their original weight needs to be multiplied by `coef` defined below
            coef = (100.0 - non_trigger_pct) / 100.0
            for k in expt_variant_weights.keys():
                if k != CATEG_NAN_VALUE:
                    expt_variant_weights[k] = expt_variant_weights[k] * coef
        experiment_choices = list(expt_variant_weights.keys())
        experiment_probs = list(expt_variant_weights.values())
        expt_df[VARIANT_COL] = rng.choice(experiment_choices, p=experiment_probs, size=sample_size)

        summary = expt_df[VARIANT_COL].value_counts(normalize=True).to_dict()

        return expt_df, summary

    def simulate_expt_assignment_multi(
        self,
        population_df: pd.DataFrame,
        expt_variant_weights_multi: tuple[dict[str, float]],
        population_pcnt_multi: tuple[float] = tuple(),
        non_trigger_pct_multi: tuple[float] = tuple(),
        seed_multi: tuple[int] = tuple(),
    ) -> list[tuple[pd.DataFrame, dict[str, dict[str, float]]]]:
        """
        Samples the population for multiple experiments and returns the results.

        Args:
            population_df: The generated population DataFrame.
            expt_variant_weights_multi: List of experiment weights for each experiment.
            population_pcnt_multi: Percentage of the population to sample.
            non_trigger_pct_multi: List of non-trigger percentages for each experiment.
            seed_multi: Random seed for reproducibility.

        Returns:
            result: List of sampled DataFrames and their variant summaries.
        """
        if not expt_variant_weights_multi:
            raise ValueError("`expt_variant_weights_multi` cannot be None or empty.")

        k = len(expt_variant_weights_multi)  # Infer the number of experiments (k)
        results = []

        for i in range(k):
            expt_variant_weights = expt_variant_weights_multi[i]

            population_pcnt = 100.0
            if population_pcnt_multi:
                population_pcnt = population_pcnt_multi[i]

            non_trigger_pct = 0.0
            if non_trigger_pct_multi:
                non_trigger_pct = non_trigger_pct_multi[i]

            seed = None
            if seed_multi:
                seed = seed_multi[i]

            expt_df, variant_summary = self.simulate_expt_assignment(
                population_df=population_df,
                expt_variant_weights=expt_variant_weights,
                population_pcnt=population_pcnt,
                non_trigger_pct=non_trigger_pct,
                seed=seed,
            )
            results.append((expt_df, variant_summary))

        return results

    def merge_expt_dfs(
        self, expt_df_list: list[pd.DataFrame], attributes: tuple[str] = tuple()
    ) -> tuple[pd.DataFrame, dict]:
        """
        Merges results from multiple experiments into a single DataFrame with variant tuples.

        Args:
            expt_df_list: expt assignment dataframe from multiple experiments.
            attributes: Columns to use for merging.

        Returns:
            result: A tuple:

                - pd.DataFrame: Merged DataFrame with experiment variant tuples.
                - dict: summary

        """
        on_cols = [UNIT_COL]
        if attributes:
            on_cols += list(attributes)

        # This will join the various experiments (outer join)
        # And generate multi-expt variants as tuples
        # It will also replace Nan generated during outer join with CATEG_NAN_VALUE
        dc_list = [DataContainer(df=df, is_df=True) for df in expt_df_list]
        expt_dc = join_expt_dfs(dc_list=dc_list, on_cols=on_cols, common_cols=[VARIANT_COL])
        expt_df = expt_dc.df

        summary = expt_df[VARIANT_COL].value_counts(normalize=True).to_dict()

        return expt_df, summary

    def apply_experiment_impact(
        self,
        df: pd.DataFrame,
        expt_metric_impacts: tuple[dict[str, float]],
        interaction_metric_impacts: dict[tuple[str], dict[str, float]] = dict(),
    ) -> pd.DataFrame:
        """
        Applies experiment impacts to metrics and includes interaction effects.

        Args:
            df: Merged DataFrame with variant tuples.
            expt_metric_impacts: Experiment impacts on metrics.
                The tuple is assumed to be ordered with k-th element which is a dict
                corresponding to eperiment k. See example.

            expt_metric_impacts = [
                {"control": {"metric1": 5}, "v1": {"metric1": -3}, "v2": {"metric1": +6}},  # Experiment 1 impact
                {"control": {"metric2": 10}, "enabled": {"metric2": 12}},  # Experiment 2 impact
            ]

            interaction_metric_impacts: Interaction impacts based on variant tuples.

        Returns:
            result: Updated DataFrame with applied impacts.
        """
        if expt_metric_impacts:
            for i, variant_impacts in enumerate(expt_metric_impacts):
                variant_col = f"{VARIANT_COL}_{i + 1}"
                for variant, impacts in variant_impacts.items():
                    for metric_name, impact_value in impacts.items():
                        df.loc[df[variant_col] == variant, metric_name] += impact_value

        if interaction_metric_impacts:
            for variant, impacts in interaction_metric_impacts.items():
                for metric_name, impact_value in impacts.items():
                    df.loc[df[VARIANT_COL] == variant, metric_name] += impact_value

        return df

    def run(self):
        """
        Runs the simulation end to end and updates the class attributes.
        """
        # Generate a population
        self.population_df, self.population_summary = self.generate_population(
            population_size=self.population_size,
            attribute_weights=self.attribute_weights,
            seed=self.population_seed,
        )

        # Sample population and run multiple experiments
        self.expt_data_list = self.simulate_expt_assignment_multi(
            population_df=self.population_df,
            expt_variant_weights_multi=self.expt_variant_weights_multi,
            population_pcnt_multi=self.population_pcnt_multi,
            non_trigger_pct_multi=self.non_trigger_pct_multi,
            seed_multi=self.expt_assignment_seed_multi,
        )

        expt_df_list = [data[0].copy() for data in self.expt_data_list]

        # Merge experiment results
        self.expt_df, self.expt_df_summary = self.merge_expt_dfs(
            expt_df_list=expt_df_list, attributes=(self.attribute_weights.keys())
        )

        # We can metrics only if metric attributes are passed.
        if self.metric_attribute_values:
            self.expt_metric_baseline_df = self.add_baseline_metrics(
                df=self.expt_df.copy(), metric_attribute_values=self.metric_attribute_values
            )

            # Apply metrics and interaction effects
            self.expt_metric_df = self.apply_experiment_impact(
                df=self.expt_metric_baseline_df.copy(),
                expt_metric_impacts=self.expt_metric_impacts,
                interaction_metric_impacts=self.interaction_metric_impacts,
            )

    def publish(self, file: Optional[TextIO] = None) -> None:
        """
        Prints information about the simulation. If an open file is passed
        then it writes the results into file.

        Args:
            file: An open file or None.
        """
        print(f"\nPopulation Summary:\n{self.population_summary}", file=file)
        print(f"\nsim.population_df:\n{self.population_df}", file=file)

        print("\nexpt_data_list (one by one):\n", file=file)
        for i, expt_data in enumerate(self.expt_data_list):
            expt_df = expt_data[0]
            summary = expt_data[1]
            print(f"\nexpt_df {i}:\n{expt_df}", file=file)
            print(f"\nsummary {i}:\n{summary}", file=file)
            print(f"\nexpt_df.shape {i}:\n{expt_df.shape}", file=file)

        print(f"\nmerged expt df (sim.expt_df):\n{self.expt_df}", file=file)
        print(f"\nsim.expt_df.shape:\n{self.expt_df.shape}", file=file)

        print(f"\nsim.expt_df_summary:\n{self.expt_df_summary}", file=file)

        if self.metric_attribute_values:
            print(
                f"\nbaseline expt metric df, sim.expt_metric_baseline_df:\n{self.expt_metric_baseline_df.head()}",
                file=file,
            )
            print(f"\nsim.expt_metric_df.shape:\n{self.expt_metric_baseline_df.shape}", file=file)

            agg_dict = {metric: "mean" for metric in self.metric_attribute_values}

            attributes = list(self.attribute_weights.keys())

            mean_attribute_metric_baseline_df = self.expt_metric_baseline_df.groupby(
                attributes
            ).agg(agg_dict)
            print(
                f"\nmean_attribute_metric_baseline_df:\n{mean_attribute_metric_baseline_df}",
                file=file,
            )

            mean_variant_metric_baseline_df = self.expt_metric_baseline_df.groupby(
                [VARIANT_COL]
            ).agg(agg_dict)
            print(f"\nmean_variant_metric_df:\n{mean_variant_metric_baseline_df}", file=file)

            print(
                f"\ndf with Impacts, sim.expt_metric_df:\n{self.expt_metric_df.head()}", file=file
            )
            print(f"\nsim.expt_metric_df.shape:\n{self.expt_metric_df.shape}", file=file)

            mean_attribute_metric_df = self.expt_metric_df.groupby(attributes).agg(agg_dict)
            print(f"\nmmean_attribute_metric_df:\n{mean_attribute_metric_df}", file=file)

            mean_variant_metric_df = self.expt_metric_df.groupby([VARIANT_COL]).agg(agg_dict)
            print(f"\nmean_variant_metric_df:\n{mean_variant_metric_df}", file=file)

            mean_attribute_variant_metric_df = self.expt_metric_df.groupby(
                attributes + [VARIANT_COL]
            ).agg(agg_dict)
            print(
                f"\nmean_attribute_variant_metric_df:\n{mean_attribute_variant_metric_df}",
                file=file,
            )



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini


import os
from dataclasses import dataclass
from typing import List, Optional

import pandas as pd
import plotly.graph_objects as go

from abvelocity.param.io_param import IOParam
from abvelocity.utils.get_categ_column_distbn import get_categ_column_distbn


@dataclass
class BarchartResult:
    """Dataclass to hold the results of barchart generation."""

    fig: go.Figure
    df: pd.DataFrame
    plot_df: pd.DataFrame
    query: str
    num_distict_labels: int


class JourneyBarchart:
    """Generates bar charts for event sequences."""

    def __init__(self, io_param: IOParam):
        """
        Initializes the JourneyBarchart instance.

        Args:
            io_param (IOParam): Dataclass containing input/output parameters.
        """
        self.cursor = io_param.cursor
        self.print_to_html = io_param.print_to_html
        self.save_path = io_param.save_path
        self.file_name_suffix = io_param.file_name_suffix
        os.makedirs(self.save_path, exist_ok=True)

    def gen(
        self,
        table_name: str,
        count_distinct_col: Optional[str] = None,
        col: str = "event_seq",
        max_num: int = 25,
        conditions: List[str] = ["TRUE = TRUE"],
        sort_event_array: bool = False,
        is_map: bool = False,
        title: Optional[str] = None,
    ) -> BarchartResult:
        """
        Generates a bar chart for event sequences.

        Args:
            table_name (str): Name of the table containing the sequence data.
            count_distinct_col (Optional[str]): Column to count distinct values from.
            col (str): Column containing the event sequences.
            max_num (int): Maximum number of bars in the chart.
            conditions (List[str]): List of SQL conditions to filter data.
            sort_event_array (bool): Whether to sort events within sequences.
                This is not enabled for maps as conversions will be too complex.
            is_map (bool): Whether the seq is a map or not.
            title (str, optional): Optional title. One will be constructed based on input, if not passed.

        Returns:
            BarchartResult: Dataclass containing the generated chart data.
        """

        if is_map:
            table_name2 = f"""
                (SELECT
                    JSON_FORMAT(CAST(event_seq AS JSON)) AS event_seq_str,
                    *
                FROM
                {table_name}
                )
            """
        else:
            table_name2 = f"""
                (SELECT
                    ARRAY_JOIN({col}, ',') AS event_seq_str,
                    *
                FROM
                    {table_name}
                )
            """

            if sort_event_array:
                table_name2 = f"""
                    (SELECT
                        ARRAY_JOIN(ARRAY_SORT({col}), ',') AS event_seq_str,
                        *
                    FROM
                        {table_name}
                    )
                """

        res_dict = get_categ_column_distbn(
            table_name=table_name2,
            col="event_seq_str",
            conditions=conditions,
            max_num=max_num,
            cursor=self.cursor,
            print_to_html=None,  # We handle the printing to html below
            title=title,
        )

        file_name = os.path.join(
            self.save_path, f"barchart{self.file_name_suffix}_{table_name.replace('.', '_')}.html"
        )
        if self.print_to_html:
            self.print_to_html(res_dict["fig"], file_name=file_name)

        return BarchartResult(**res_dict)



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini
"""This module includes dataclasses to for parameters used in analysis of experiments."""

from dataclasses import dataclass
from typing import Optional

from abvelocity.param.metric import Metric
from abvelocity.param.metric_family import MetricFamily


@dataclass
class MetricInfo:
    """
    This is a dataclass to contain the information for a group of metrics.
    """

    metric_family: Optional[MetricFamily] = None
    """
    The metric family for the analysis.

        - This essentially says which table the metrics belong too and includes extra
        processing functions if needed (this feature should be rarely needed).

        - It also might include `metrics` field available which can be used
        to fill in `metrics` field in this dataclass.

    """
    metrics: Optional[list[Metric]] = None
    """
    metrics to be investigated.
    This might be inferred from `metric_family` if this field is not passed.
    """
    start_date: Optional[str] = None
    """This is the analysis start data."""
    end_date: Optional[str] = None
    """This is the analysis end data."""

    def __post_init__(self):
        """Infers `metrics` from `metric_family` if available."""
        # Checks if `self.metrics` is None but `self.metric_family.metrics` is avaliable.
        if not self.metrics and self.metric_family and self.metric_family.metrics:
            self.metrics = self.metric_family.metrics

    def __str__(self):
        metrics_str = (
            "\n" + ", ".join(str(metric) for metric in self.metrics) + "\n"
            if self.metrics
            else "None"
        )

        metric_family_str = None
        if self.metric_family:
            self.metric_family_str = self.metric_family.name

        return (
            # f"Analysis Information:\n"
            # f"{self.multi_expt_info}\n"
            f"- Metrics: {metrics_str}\n"
            f"- Metric Family: {metric_family_str}"
            # f"- Analysis start date: {self.start_date}"
            # f"- Analysis end date: {self.end_date}"
        )

    def to_html(self):
        """
        Convert the MetricInfo data to a html format.
        """
        metrics_html = (
            "<p>" + ",<p> ".join(str(metric) for metric in self.metrics) + "<p>"
            if self.metrics
            else "None"
        )

        metric_family_html = "None"
        if self.metric_family and self.metric_family.name:
            self.metric_family_str = self.metric_family.name

        return (
            f"<div>"
            # f"<h3>Analysis Information</h3>"
            # f"{self.multi_expt_info.to_html()}"
            f"<p><strong>Metric Family:</strong> {metric_family_html}</p>"
            f"<p><strong>Metrics:</strong> {metrics_html}</p>"
            # f"<p><strong>Analysis Start Date:</strong> {self.start_date}</p>"
            # f"<p><strong>Analysis End Date:</strong> {self.end_date}</p>"
            f"</div>"
        )

    def to_markdown(self) -> str:
        """
        Convert the MetricInfo data to a Markdown format.
        """
        metrics_md = (
            "\n" + ", ".join(str(metric) for metric in self.metrics) + "\n"
            if self.metrics
            else "None"
        )

        metric_family_md = "None"
        if self.metric_family and self.metric_family.name:
            metric_family_md = self.metric_family.name

        return f"**Metric Family**: {metric_family_md}\n" f"**Metrics**: {metrics_md}\n"



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini


from dataclasses import dataclass, field
from typing import List, Optional

CONVERSION_RATE = "conversion_rate"
"""The default name for the conversion rate column."""

NUMER_COUNT = "numer_count"
"""The default name for the numerator of conversion rate computation."""

DENOM_COUNT = "denom_count"
"""Th edefault for the denominator of conversion rate computation."""


@dataclass
class ConversionQuery:
    """
    A class to generate SQL queries for calculating conversion rates
    based on the presence of specific values within an array column.

    The conversion is calculated as:
        (Count of entities where `array_col` contains specified numerator values) /
        (Count of entities where `array_col` contains specified denominator values)

    If `numerator_list` or `denominator_list` is None, all rows are considered
    to satisfy that respective condition.

    `require_all_numerator` and `require_all_denominator` control whether *all*
    elements from the respective list must be present in `array_col` (AND logic)
    or *any* of them (OR logic, default).

    Counts can be distinct on a `count_distinct_col` (e.g., user_id) or simple row counts.

    Attributes:
        table_name (str): The name of the table containing the data.
            This can be a base query like `(SELECT * FROM my_table WHERE ...)`
        array_col (str): The name of the array column to analyze (e.g., 'event_tags').
        numerator_list (Optional[List[str]]): A list of values in the array_col
            that define the numerator (e.g., ['product_view']).
            If None, all rows are included in the numerator count.
        denominator_list (Optional[List[str]]): A list of values in the array_col
            that define the denominator (e.g., ['page_load']).
            If None, all rows are included in the denominator count.
        require_all_numerator (bool): If True, all values in `numerator_list` must be
            present in `array_col`. If False, any value is sufficient. Defaults to False.
        require_all_denominator (bool): If True, all values in `denominator_list` must be
            present in `array_col`. If False, any value is sufficient. Defaults to False.
        count_distinct_col (Optional[str]): The column to use for COUNT(DISTINCT).
            If provided, counts unique entities (e.g., 'user_id').
            If None, counts rows satisfying the conditions.
        conditions (Optional[List[str]]): A list of individual SQL WHERE conditions
            (e.g., ["event_date >= DATE '2023-01-01'", "device_type = 'mobile'"]).
            These will be joined by ' AND '.
        group_by_cols (Optional[List[str]]): A list of columns to group the results by
            (e.g., ['DATE(event_timestamp)', 'country']).
    """

    table_name: str
    """The name of the table containing the data."""

    array_col: str
    """The name of the array column containing the events/tags."""

    numerator_list: Optional[List[str]]
    """A list of values in the array_col that define the numerator, or None to count all rows."""

    denominator_list: Optional[List[str]]
    """A list of values in the array_col that define the denominator, or None to count all rows."""

    require_all_numerator: bool = False
    """If True, all values in `numerator_list` must be present. If False, any is sufficient."""

    require_all_denominator: bool = False
    """If True, all values in `denominator_list` must be present. If False, any is sufficient."""

    count_distinct_col: Optional[str] = None
    """The column to use for COUNT(DISTINCT).
    If provided, counts unique entities (e.g., 'user_id').
    If None, counts rows satisfying the conditions.
    """

    # Using default_factory for mutable defaults like lists is a dataclass best practice
    conditions: Optional[List[str]] = field(default_factory=list)
    """A list of individual SQL WHERE conditions (e.g., ["event_date >= DATE '2023-01-01'", "device_type = 'mobile'"]).
    These will be joined by ' AND '.
    """

    group_by_cols: Optional[List[str]] = field(default_factory=list)
    """A list of columns to group the results by (e.g., ['DATE(event_timestamp)', 'country'])."""

    col_name: Optional[str] = CONVERSION_RATE
    """The name for the conversion rate column."""

    def gen(self) -> str:
        """
        Generates a Presto SQL query string to calculate a conversion rate based
        on the parameters provided during the object's initialization.
        The query now also includes the raw numerator and denominator counts.

        Returns:
            str: The generated Presto SQL query.
        """

        # Validate numerator_list
        if self.numerator_list is not None and (
            not isinstance(self.numerator_list, list) or not self.numerator_list
        ):
            raise ValueError("numerator_list must be a non-empty list or None.")
        # Validate denominator_list
        if self.denominator_list is not None and (
            not isinstance(self.denominator_list, list) or not self.denominator_list
        ):
            raise ValueError("denominator_list must be a non-empty list or None.")

        if self.count_distinct_col is not None and not isinstance(self.count_distinct_col, str):
            raise TypeError("count_distinct_col must be a string or None.")
        if not isinstance(self.conditions, list) or not all(
            isinstance(cond, str) for cond in self.conditions
        ):
            raise TypeError("conditions must be a list of strings or None.")
        if not isinstance(self.group_by_cols, list) or not all(
            isinstance(col, str) for col in self.group_by_cols
        ):
            raise TypeError("group_by_cols must be a list of strings or None.")

        def _get_array_condition(event_list: Optional[List[str]], require_all: bool) -> str:
            """
            Helper function to generate the SQL condition for array containment.
            """
            if event_list is None:
                return "TRUE"
            if not event_list:
                # This case is already handled by outer validation, but defensive check
                raise ValueError("List cannot be empty if not None.")

            # Define single quote characters outside the f-string expression
            _s_quote = "'"
            _d_s_quote = "''"

            # Prepare the SQL list string for ARRAY function
            sql_list_items = ", ".join(
                [
                    (
                        f"{_s_quote}{item.replace(_s_quote, _d_s_quote)}{_s_quote}"
                        if isinstance(item, str)
                        else str(item)
                    )
                    for item in event_list
                ]
            )

            # Condition using ARRAY_INTERSECT for both 'any' and 'all' logic
            array_intersect_clause = (
                f"CARDINALITY(ARRAY_INTERSECT({self.array_col}, ARRAY[{sql_list_items}]))"
            )

            if require_all:
                # For 'all' logic, cardinality of intersection must equal the length of the list
                return f"{array_intersect_clause} = {len(event_list)}"
            else:
                # For 'any' logic, cardinality of intersection must be greater than 0
                return f"{array_intersect_clause} > 0"

        numerator_array_condition = _get_array_condition(
            self.numerator_list, self.require_all_numerator
        )
        denominator_array_condition = _get_array_condition(
            self.denominator_list, self.require_all_denominator
        )

        count_expr_numerator = ""
        count_expr_denominator = ""

        if self.count_distinct_col:
            count_expr_numerator = (
                f"COUNT(DISTINCT CASE WHEN {numerator_array_condition} "
                f"THEN {self.count_distinct_col} ELSE NULL END)"
            )
            count_expr_denominator = (
                f"COUNT(DISTINCT CASE WHEN {denominator_array_condition} "
                f"THEN {self.count_distinct_col} ELSE NULL END)"
            )
        else:
            count_expr_numerator = (
                f"COUNT(CASE WHEN {numerator_array_condition} THEN 1 ELSE NULL END)"
            )
            count_expr_denominator = (
                f"COUNT(CASE WHEN {denominator_array_condition} THEN 1 ELSE NULL END)"
            )

        select_parts = []
        if self.group_by_cols:
            select_parts.extend(self.group_by_cols)

        # Add numerator and denominator counts to the SELECT statement
        select_parts.append(f"{count_expr_numerator} AS {NUMER_COUNT}")
        select_parts.append(f"{count_expr_denominator} AS {DENOM_COUNT}")

        # Calculate conversion rate, handling division by zero by returning NULL
        conversion_expression = (
            f"CAST({count_expr_numerator} AS DOUBLE) / NULLIF({count_expr_denominator}, 0)"
        )
        select_parts.append(f"{conversion_expression} AS {self.col_name}")

        query = f"SELECT {', '.join(select_parts)} FROM {self.table_name}"

        if self.conditions:
            query += f" WHERE {' AND '.join(self.conditions)}"

        if self.group_by_cols:
            query += f" GROUP BY {', '.join(self.group_by_cols)}"

        self.query = query.strip()

        return self.query



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# Original author: Reza Hosseini

import datetime
import os
import warnings
from dataclasses import asdict, dataclass
from typing import Optional

import pandas as pd

from abvelocity.get_data.get_expt_stats import get_expt_stats
from abvelocity.param.analysis_info import AnalysisInfo
from abvelocity.param.constants import (
    CATEG_NAN_VALUE,
    CI_COL,
    CI_PERCENT_COL,
    CONTROL_LABEL,
    DELTA_SUM_CI_COL,
    METRIC_NAME_COL,
    TRIGGER_STATE_COUNT_COL,
    VARIANT_COL,
)
from abvelocity.param.launch import Launch
from abvelocity.param.launch_to_comparison_pair import launch_to_comparison_pair
from abvelocity.param.variant import ComparisonPair
from abvelocity.stats.calc_variant_metric_effects import (
    calc_variant_metric_effects,
    compare_variants_with_control,
)
from abvelocity.stats.calc_variant_metric_stats import calc_variant_metric_stats
from abvelocity.stats.two_sample_z_test import two_sample_z_test
from abvelocity.utils.calc_freq import calc_freq
from abvelocity.utils.df_to_html import df_to_html
from abvelocity.utils.round_df import round_df


@dataclass
class MEAMetricResult:
    """This dataclass stores MEA (Multi-Expt Analysis) results for one metric."""

    variant_metric_stats_df: Optional[pd.DataFrame] = None
    """Variant statistics dataframe."""
    variant_effect_df_pairs: Optional[pd.DataFrame] = None
    """Variant effects dataframe for comparison pairs."""
    variant_effect_df_pairs_sig: Optional[pd.DataFrame] = None
    """Variant effects dataframe for comparison pairs for sig results only."""
    variant_effect_df_vs_control: Optional[pd.DataFrame] = None
    """Variant effects dataframe for all variants versus control."""
    univar_variant_metric_stats_df: Optional[pd.DataFrame] = None
    """Univariate variant statistics dataframe."""
    univar_variant_effect_df_vs_control: Optional[pd.DataFrame] = None
    """Univariate variant effects dataframe for all variants versus control."""


# END_USER_METRIC_RESULT_KEYS = ["variant_effect_df_pairs", "variant_effect_df_pairs_sig"]
END_USER_METRIC_RESULT_KEYS = ["variant_effect_df_pairs_sig"]
"""The end user does not need all the genarted tables.
This constant is used to decide what is to be shown in the final report.
Currently we only include the comparisons.
These could be launch effects for example passed using `Launch` and then converted to `ComparisonPair`.
"""

END_USER_COLS = [
    METRIC_NAME_COL,
    "launch",
    "delta",
    "delta_percent",
    CI_PERCENT_COL,
    "p_value",
    "delta_sum",
]
"""
The columns to include in metric effect tables for end user.
"""


@dataclass
class MEAResult:
    """This dataclass stores the results of MEA (Multi-Expt Analysis).
    The results include variant frequencies and MEA results for each metric.
    """

    variant_freq_dict: Optional[dict[str, pd.DataFrame]] = None
    """Dictionary of variant assignment frequencies.
        The keys are:
            - expt_1
            - expt_2
            ...
            - expt_k (k is the number of experiments)
            - multi-expt
    """
    metric_result_dict: Optional[dict[str, MEAMetricResult]] = None
    """Dictionary of MEA results for each metric.
        The keys are the metric names.
    """
    combined_mea_result: Optional[MEAMetricResult] = None
    """
    This is an `MEAMetricResult` which includes all metrics in its corresponding tables.
    This is done by concatenating all the dataframes in individual metric results.
    To each dataframe a new "metric" column added and then data are concatenated
    vertically.
    """

    def gen_combined_mea_result(self) -> None:
        """
        This creates an `MEAMetricResult` which includes all metrics in its corresponding tables.
            This is done by concatenating all the dataframes in individual metric results.
            To each dataframe a new "metric" column added and then data are concatenated
            vertically.

        Alters: `self.combined_mea_result`
        """
        self.combined_mea_result = MEAMetricResult()
        # We only combine the key tables.
        variant_metric_stats_df = None
        variant_effect_df_pairs = None
        # add a metric  name column and concat al dataframes.
        for metric, mea_metric_result in self.metric_result_dict.items():
            variant_metric_stats_df0 = mea_metric_result.variant_metric_stats_df
            if variant_metric_stats_df0 is not None:
                variant_metric_stats_df0[METRIC_NAME_COL] = metric
                cols = list(variant_metric_stats_df0.columns)
                # re-arrange the columns
                cols = [cols[-1]] + cols[:-1]
                variant_metric_stats_df0 = variant_metric_stats_df0[cols]

                variant_metric_stats_df = pd.concat(
                    [variant_metric_stats_df0, variant_metric_stats_df], axis=0
                )
            variant_effect_df_pairs0 = mea_metric_result.variant_effect_df_pairs
            if variant_effect_df_pairs0 is not None:
                variant_effect_df_pairs0[METRIC_NAME_COL] = metric
                cols = list(variant_effect_df_pairs0.columns)
                # re-arrange the columns
                cols = [cols[-1]] + cols[:-1]
                variant_effect_df_pairs0 = variant_effect_df_pairs0[cols]

                variant_effect_df_pairs = pd.concat(
                    [variant_effect_df_pairs0, variant_effect_df_pairs], axis=0
                )

        variant_effect_df_pairs_sig = variant_effect_df_pairs.loc[
            variant_effect_df_pairs["p_value"] < 0.10
        ]
        self.combined_mea_result.variant_metric_stats_df = variant_metric_stats_df
        self.combined_mea_result.variant_effect_df_pairs = variant_effect_df_pairs
        self.combined_mea_result.variant_effect_df_pairs_sig = (
            variant_effect_df_pairs_sig.reset_index(drop=True)
        )

    def combine(self, other):
        """
        This function will combine the results from two mea runs.
        For dict objects it will update them using other.

        For the dataframes given in `combined_mea_result`, it will concat them
        vertically.
        """
        if self.variant_freq_dict and other.variant_freq_dict:
            self.variant_freq_dict.update(other.variant_freq_dict)
        elif other.variant_freq_dict:
            self.variant_freq_dict = other.variant_freq_dict

        if self.metric_result_dict and other.metric_result_dict:
            self.metric_result_dict.update(other.metric_result_dict)
        elif other.metric_result_dict:
            self.metric_result_dict = other.metric_result_dict

        if self.combined_mea_result and other.combined_mea_result:
            field_names = [
                "variant_metric_stats_df",
                "variant_effect_df_pairs",
                "variant_effect_df_pairs_sig",
            ]
            for field_name in field_names:
                res1 = getattr(self.combined_mea_result, field_name)
                res2 = getattr(other.combined_mea_result, field_name)

                if res1 is not None and res2 is not None:
                    res = pd.concat([res1, res2], ignore_index=True)
                elif res2 is not None:
                    res = res2

                setattr(self.combined_mea_result, field_name, res)
        elif other.combined_mea_result:
            self.combined_mea_result = other.combined_mea_result


@dataclass
class MEAReport:
    """This encodes the MEA report.
    The main component is `html_str` which can be published.
    """

    html_str: str = ""
    """
    html string which can be stored in an html file.
    """
    file_names: Optional[list] = None
    """
    file names generated during publish, if any.
    """
    paths: Optional[list] = None
    """
    paths generated during publish if any.
    """


class MEA:
    """This class is used to run multi-experiment analysis (MEA).
    The main method is `run` which generates the results (See `~abvelocity.mea.mea.MEAResult`).
    The class also includes a `publish` method to generate html string and write results to csv files.
    """

    def __init__(
        self,
        df: Optional[pd.DataFrame] = None,
        analysis_info: Optional[AnalysisInfo] = None,
        launches: Optional[list[Launch]] = None,
        control_launch: Optional[Launch] = None,
        comparison_pairs: Optional[list[ComparisonPair]] = None,
        ci_coverage: float = 0.95,
        recalculate_expt_stats: bool = False,
    ):
        """Initializes the class with the data, analysis info and comparison pairs.

        Args:
            df: Dataframe with the data.
            analysis_info: AnalysisInfo with the experiments and metrics.
            launches: List of launches.
                Each launch is a combination of variants across experiments
                (or simply a string for single experiment case).
                See `~abvelocity.param.launch.Launch`.
                For each launch first,

                    - we construct the counterpart multi-experiment control
                    - for each arm we map them to the corresponding `VariantList`.
                    - a `ComparisonPair` is created with these two objects then.

                All the above steps are done using `launch_to_comparison_pair` function.
            control_launch: The Launch which is used as the baseline for comparison.
                If not passed, we will assume all experiments are on the control arm (`CONTROL_LABEL`).
            comparison_pairs: List of comparison pairs.
                This is an advanced feature as `launches` will construct the needed comparison_pairs for typical launches.
                Each pair includes a treatment and control field.
                See `~abvelocity.param.variant.ComparisonPair`.
            ci_coverage: coverage of conf interval.
            recalculate_expt_stats: A bool to decide if experiment derived stats should be recalculated.

        Attributes (other than the input arguments):
            mea_result: Result of MEA. This is attached after running the analysis.
        """
        self.df = df
        self.analysis_info = analysis_info
        self.launches = launches
        self.control_launch = control_launch
        self.comparison_pairs = comparison_pairs
        self.ci_coverage = ci_coverage
        self.recalculate_expt_stats = recalculate_expt_stats

        # Attributes
        self.result = None

    def run(self) -> None:
        """
        This is the main method for the class.
        It runs multi-experiment analysis (MEA).
        Note that the arguments (See `__init__`) can be passed during the class initiation or
        attached later.

        Args:
            None. (Uses attributes of `self`, see `__init__`).

        Alters:
            self: attches the result with type `MEAResult` to the class instance.

        """
        df = self.df
        analysis_info = self.analysis_info
        launches = self.launches
        control_launch = self.control_launch
        comparison_pairs = self.comparison_pairs
        expt_info_list = analysis_info.multi_expt_info.expt_info_list

        if (not analysis_info.multi_expt_info.derived_stats) or self.recalculate_expt_stats:
            analysis_info.multi_expt_info.derived_stats = get_expt_stats(df=df)

        if not launches:
            # If no `control_launch` is specified, we simply get the non control launches.
            # These are launches for which at least one launch is not on control arm.
            if control_launch is None:
                launches = analysis_info.multi_expt_info.derived_stats.non_control_launches
            else:
                # Get all launches
                launches = analysis_info.multi_expt_info.derived_stats.launches
                # Remove the spcified `control_launch`.
                # The way to do that is to check all launches and compare with `control_launch`.
                # This is done using the equality defined for `Variant` dataclass.
                # This equality only depends on the value field and not the name.
                launches = [launch for launch in launches if launch.value != control_launch.value]

            print("\n*** launches were inferred from derived_stats:" f"\n{launches}")

        # We will convert None to empty list, as we might add new comparison pairs if `launches` is not None.
        if not comparison_pairs:
            comparison_pairs = []

        if launches:
            for launch in launches:
                comparison_pair = launch_to_comparison_pair(
                    launch=launch, control_launch=control_launch
                )
                comparison_pairs.append(comparison_pair)

        if not analysis_info.metric_info_list:
            self.result = None
            warnings.warn(
                f"\n***: metric_info_list is None or empty = {analysis_info.metric_info_list}"
                "This means MEA will only generate overlap results.",
                UserWarning,
            )
            return None

        metrics = []
        for metric_info in analysis_info.metric_info_list:
            if metric_info.metrics is None:
                raise ValueError(
                    "At this stage metrics need to be available."
                    f"Check analysis_info: {analysis_info}"
                    f"Check metric_info: {metric_info}"
                )
            metrics += metric_info.metrics

        print(f"\n *** MEA: metrics:\n{metrics}")
        num_expts = len(expt_info_list)

        print(f"\n*** MEA: df:\n{df.head(2)}")

        # Get stats about experiment variant assignments
        variant_freq_dict = {}
        for i in range(num_expts):
            col = f"{VARIANT_COL}_{i + 1}"
            variant_freq_dict[f"expt_{i + 1}"] = calc_freq(
                df=df[df[col] != CATEG_NAN_VALUE], cols=[col]
            )

        variant_freq_dict["multi-expt"] = calc_freq(df=df, cols=[VARIANT_COL])

        metric_result_dict = {}
        expt_control = (CONTROL_LABEL,) * num_expts
        print(f"\n*** MEA: expt_control:\n{expt_control}")

        # Get stats about the multi-experiment
        variant_count_df = analysis_info.multi_expt_info.derived_stats.variant_count_df

        for metric in metrics:
            print(f"\n**** MEA for metric name: {metric.name}")
            print(f"\n**** MEA for metric def: {metric}")
            metric_result_dict[metric.name] = MEAMetricResult()
            # TODO: extend to more general metrics (allow denominators)
            # Currently it works only for metrics with a numerator.
            variant_metric_stats_df = calc_variant_metric_stats(
                df=df, metric=metric, variant_col=VARIANT_COL
            )

            variant_effect_df_pairs = None
            if comparison_pairs:  # this will make sure its not None or empty.
                variant_effect_df_pairs = calc_variant_metric_effects(
                    variant_metric_stats_df=variant_metric_stats_df,
                    variant_col=VARIANT_COL,
                    comparison_pairs=comparison_pairs,
                    stats_test_func=two_sample_z_test,
                    ci_coverage=self.ci_coverage,
                    variant_count_df=variant_count_df,
                    trigger_state_count_col=TRIGGER_STATE_COUNT_COL,
                )

            variant_effect_df_vs_control = compare_variants_with_control(
                variant_metric_stats_df=variant_metric_stats_df,
                variant_col=VARIANT_COL,
                expt_control=expt_control,
                stats_test_func=two_sample_z_test,
                ci_coverage=self.ci_coverage,
                variant_count_df=variant_count_df,
                trigger_state_count_col=TRIGGER_STATE_COUNT_COL,
            )

            metric_result_dict[metric.name].variant_metric_stats_df = variant_metric_stats_df
            metric_result_dict[metric.name].variant_effect_df_pairs = variant_effect_df_pairs
            metric_result_dict[metric.name].variant_effect_df_vs_control = (
                variant_effect_df_vs_control
            )

        # Univar results
        for metric in metrics:
            univar_variant_metric_stats_df = None
            univar_variant_effect_df_vs_control = None
            for i in range(num_expts):
                # TODO: extend to more general metrics (allow denominators)
                variant_metric_stats_df = calc_variant_metric_stats(
                    df=df, metric=metric, variant_col=f"{VARIANT_COL}_{i + 1}"
                )
                cols = list(variant_metric_stats_df.columns)
                # re-arrange the columns
                cols = [cols[-1]] + cols[:-1]
                variant_metric_stats_df = variant_metric_stats_df[cols]
                # Change the variant column name from `f"{VARIANT_COL}_{i + 1}"` to `VARIANT_COL`.
                # This is needed for merging univariate data across experiments.
                variant_metric_stats_df.rename(
                    columns={f"{VARIANT_COL}_{i + 1}": VARIANT_COL}, inplace=True
                )

                univar_variant_metric_stats_df = pd.concat(
                    [univar_variant_metric_stats_df, variant_metric_stats_df], axis=0
                )

                if expt_info_list[i].derived_stats:
                    variant_count_df_uni = expt_info_list[i].derived_stats.variant_count_df
                else:
                    variant_count_df_uni = None

                variant_effect_df_vs_control = compare_variants_with_control(
                    variant_metric_stats_df=variant_metric_stats_df,
                    variant_col=VARIANT_COL,
                    expt_control=CONTROL_LABEL,
                    stats_test_func=two_sample_z_test,
                    ci_coverage=self.ci_coverage,
                    variant_count_df=variant_count_df_uni,
                    trigger_state_count_col=TRIGGER_STATE_COUNT_COL,
                )

                variant_effect_df_vs_control["expt"] = f"expt_{i}"

                univar_variant_effect_df_vs_control = pd.concat(
                    [univar_variant_effect_df_vs_control, variant_effect_df_vs_control], axis=0
                )
                cols = list(univar_variant_effect_df_vs_control.columns)
                # Reshuffles column order.
                cols = [cols[-1]] + cols[:-1]
                univar_variant_effect_df_vs_control = univar_variant_effect_df_vs_control[cols]

                metric_result_dict[metric.name].univar_variant_metric_stats_df = (
                    univar_variant_metric_stats_df
                )
                metric_result_dict[metric.name].univar_variant_effect_df_vs_control = (
                    univar_variant_effect_df_vs_control
                )

        # Attach the results to the class instance.
        self.result = MEAResult(
            variant_freq_dict=variant_freq_dict, metric_result_dict=metric_result_dict
        )

        # Generate the combined metrics results (one table for all metrics for each quantity)
        self.result.gen_combined_mea_result()

        return None

    def publish_df_color_code_p_value(
        self,
        df: pd.DataFrame,
        bg_cols: tuple[str],
        df_name: str = "",
        html_str: str = "",
        markdown_str: str = "",
        split_col: Optional[str] = None,
    ):
        """
        Publishes a DataFrame with color-coded values for statistical significance and
        returns the corresponding HTML and Markdown representations.

        This function color-codes cells in the DataFrame based on their `p_value` and `delta_percent`.
        Rows with a `p_value` below 0.05 are highlighted: green for positive `delta_percent` and
        red for negative `delta_percent`. The DataFrame can be split and displayed by a specified column.

        Args:
            df (pd.DataFrame): The DataFrame to process.
            bg_cols (tuple[str]): A tuple of column names to use for background coloring.
            df_name (str, optional): A name for the DataFrame, used in the output captions. Default is an empty string.
            html_str (str, optional): Initial HTML string to append the DataFrame's representation. Default is an empty string.
            markdown_str (str, optional): Initial Markdown string to append the DataFrame's representation. Default is an empty string.
            split_col (str, optional): A column name to split the DataFrame by. If specified, separate tables are
                created for each unique value in this column. Default is None.

        Returns:
            tuple[str, str]: A tuple containing:
                - html_str (str): The resulting HTML string with the DataFrame's representation.
                - markdown_str (str): The resulting Markdown string with the DataFrame's representation.

        Notes:
            - Cells in the DataFrame are color-coded based on their `p_value` and `delta_percent` values:
              - Light green (`rgba(0, 250, 0, 0.5)`) for `p_value < 0.05` and `delta_percent >= 0`.
              - Light red (`rgba(250, 0, 0, 0.5)`) for `p_value < 0.05` and `delta_percent < 0`.
            - If `split_col` is specified, the function generates separate tables for each unique value
              in the column and appends them to the HTML and Markdown strings.
        """

        df_name_pretty = " ".join([word.capitalize() for word in df_name.split("_")])

        html_str += f"""
            <br><br><br>
            <h2 style="color: blue; font-size: 28px; margin-top: 20px; margin-bottom: 20px;">
            {df_name_pretty}
            </h2>
        """

        def func(df0, html_str, markdown_str):
            bg_colors = None
            # color coding
            if "delta_percent" in df0.columns and "p_value" in df.columns:
                bg_colors = []
                for row in df0.itertuples():
                    if row.p_value < 0.05:
                        if row.delta_percent >= 0.0:
                            bg_colors.append("rgba(0, 250, 0, 0.5)")  # Light green
                        else:
                            bg_colors.append("rgba(250, 0, 0, 0.5)")  # Light red
                    else:
                        bg_colors.append(None)
                bg_colors = tuple(bg_colors)

            html_str += df_to_html(
                df=df0,
                top_paragraphs=[],
                caption="",
                bg_colors=bg_colors,
                bg_cols=bg_cols,
            )

            markdown_str += f"""\n\n\n\n##  <font color="blue">{df_name_pretty}</font>\n\n\n\n"""
            markdown_str += "\n\n" + df0.to_markdown(index=False)

            return html_str, markdown_str

        if split_col is not None and split_col in df.columns:
            for x in df[split_col].unique():
                html_str += f"\n<h2>{df_name_pretty}: {x}</h2>\n"
                df0 = df[df[split_col] == x].reset_index(drop=True)
                html_str, markdown_str = func(df0=df0, html_str=html_str, markdown_str=markdown_str)
        else:
            html_str += f"\n<h2>{df_name_pretty}</h2>\n"
            html_str, markdown_str = func(df0=df, html_str=html_str, markdown_str=markdown_str)

        return html_str, markdown_str

    def publish(
        self,
        mea_result: Optional[MEAResult] = None,
        analysis_info: Optional[AnalysisInfo] = None,
        write_path: Optional[str] = None,
        proj_name: Optional[str] = None,
        add_timestamp_to_path: bool = True,
        rounding_digits: int = 4,
        html_file_name: Optional[str] = None,
        markdown_file_name: Optional[str] = None,
        end_user_report: bool = True,
    ) -> dict[str, str]:
        """Writes the results of MEA to a specified path in csv format.
            For each metric the results will be written in `f"{write_path}/{metric}/"`.
            If `write_path` is not passed, it will be constructed.
            Also if `add_timestamp_to_path` is True,
            a formatted timestamp is added to the path,
            to avoid overwriting existing results.

        Args:
        mea_result: MEA result
        analysis_info: Analysis info
        write_path: Path to write results. If None one such path is generated.
            The generated path is either of

            - `f"{home}/abvelocity_results/"`
            - `f"{home}/abvelocity_results/{proj_name}"`

            where `home` is the home directory of the user.
        proj_name: This is used in the path name if passed.
        add_timestamp_to_path: If True, a timestamp is added to the path. Default True.
            This is to avoid over-writing existing records.
            The format of the timestamp is `"%Y-%m-%d_%H-%M-%S"`.
        rounding_digits: Number of digits to round the results to. Default 4.
        html_file_name: If passed, the MEA results will be written to an html file.
        markdown_file_name: If passed, the MEA results will be written to markdown file.
        end_user_report: If True, only the end user report will be generated which is based
            on the tables specified in `~abvelocity.mea.mea.END_USER_METRIC_RESULT_KEYS`.

        Returns:
            results: A dictionary with keys:

                - "html_str": The constructed html string.
                - "paths": List of paths created.
                - "file_names": List of file names created.

        """
        paths = []
        file_names = []

        if not mea_result:
            mea_result = self.result

        if not analysis_info:
            analysis_info = self.analysis_info

        html_str = ""
        markdown_str = ""

        # This will add experiments information to the html report.
        if analysis_info is not None:
            html_str += analysis_info.to_html()
            markdown_str += analysis_info.to_html()

        if write_path is not None:
            paths.append(write_path)
        else:
            home = os.path.expanduser("~")
            write_path = f"{home}/abvelocity_results/"
            paths.append(write_path)
        if proj_name is not None:
            write_path = f"{write_path}/{proj_name}"
            paths.append(write_path)

        if add_timestamp_to_path:
            timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            write_path = f"{write_path}/{timestamp}"
            print(f"\n*** write_path is:\n {write_path}")
            paths.append(write_path)

        print(f"\n***: `MEA.publish`: write_path: {write_path}")
        for path in paths:
            if not os.path.exists(path):
                os.makedirs(path)
            else:
                print(f"***\n path: {path} already exists")

        if mea_result is None:
            print(
                "***\n`MEA.publish`: You do not have any MEA results for any metrics."
                "It might be because you have not passed any metrics."
                "The report still might be generated and include assignment stats."
            )
            print(f"\n*** `mea.publish`: html_file_name: {html_file_name}")
            print(f"\n*** `mea.publish`: markdown_file_name: {markdown_file_name}")
            if html_file_name is not None:
                with open(f"{write_path}/{html_file_name}", "w") as f:
                    f.write(html_str)
                    print(f"\n*** data was written to {write_path}/{html_file_name}.")

            if markdown_file_name is not None:
                with open(f"{write_path}/{markdown_file_name}", "w") as f:
                    f.write(markdown_str)
                    print(f"\n*** data was written to {write_path}/{markdown_file_name}.")

            return {
                "html_str": html_str,
                "markdown_str": markdown_str,
                "paths": paths,
                "file_names": file_names,
            }

        # metric_result_dict = mea_result.metric_result_dict
        variant_freq_dict = mea_result.variant_freq_dict

        # Define float format for writing dataframes based on `rounding_digits`.
        float_format = f"%.{rounding_digits}f"

        # Write information for the experiments.
        if not end_user_report:
            path = f"{write_path}/expt_stats/"
            if not os.path.exists(path):
                os.makedirs(path)
            for df_name, df in variant_freq_dict.items():
                if df is not None:
                    print(f"\n*** attempting to write {df_name} from variant_freq_dict")
                    file_name = f"{path}/{df_name}.csv"
                    file_names.append(file_name)
                    print(f"\n*** file_name: {file_name} being written.")

                    # Round float and tuple cols before writing data.
                    round_df(
                        df=df,
                        rounding_digits=rounding_digits,
                        tuple_cols=[CI_COL, CI_PERCENT_COL, DELTA_SUM_CI_COL],
                    )

                    df.to_csv(file_name, index=False, float_format=float_format)

                    df_name_pretty = " ".join([word.capitalize() for word in df_name.split("_")])

                    html_str += f"\n<h2>{df_name_pretty}</h2>\n"
                    html_str += df_to_html(df=df, top_paragraphs=[df_name], caption=df_name)

                    markdown_str += (
                        f"""\n\n\n\n##  <font color="blue">{df_name_pretty} </font>\n\n\n\n"""
                    )
                    markdown_str += df.to_markdown(index=False)

        # Write MEA results for each metric.
        """
        if not end_user_report:
            for metric, mea_metric_result in metric_result_dict.items():
                metric_result = asdict(mea_metric_result)
                if end_user_report:
                    metric_result = {
                        k: v for k, v in metric_result.items() if k in END_USER_METRIC_RESULT_KEYS
                    }

                path = f"{write_path}/{metric}"
                paths.append(path)
                if not os.path.exists(path):
                    os.makedirs(path)
                else:
                    print(f"\n*** path: {path} already exists")

                for df_name, df in metric_result.items():
                    if df is not None:
                        print(f"\n*** attempting to write {df_name} for {metric}")
                        print(df)
                        file_name = f"{path}/{df_name}.csv"
                        file_names.append(file_name)
                        print(f"\n*** file_name: {file_name} being written.")
                        # Round float and tuple cols before writing data.
                        round_df(
                            df=df,
                            rounding_digits=rounding_digits,
                            tuple_cols=[CI_COL, CI_PERCENT_COL, DELTA_SUM_CI_COL],
                        )

                        df.to_csv(file_name, index=False, float_format=float_format)
                        html_str += df_to_html(
                            df=df,
                            top_paragraphs=[f"metric: {metric}, {df_name}"],
                            caption=f"metric: {metric}; {df_name}",
                        )
                        markdown_str += df.to_markdown()
        """

        # Write combined tables which include all metrics in one
        for df_name, df in asdict(mea_result.combined_mea_result).items():
            if df is not None:
                print(f"\n*** attempting to write {df_name}")
                print(df)
                path = write_path
                file_name = f"{path}/all_metrics_{df_name}.csv"
                file_names.append(file_name)
                print(f"\n*** file_name: {file_name} being written.")
                if len(df) == 0:
                    print(f"\n*** df {df_name} is empty.")
                else:
                    # Round float and tuple cols before writing data.
                    round_df(
                        df=df,
                        rounding_digits=rounding_digits,
                        tuple_cols=[CI_COL, CI_PERCENT_COL, DELTA_SUM_CI_COL],
                    )

                    if "comparison_pair" in df.columns:
                        df.rename(columns={"comparison_pair": "launch"}, inplace=True)

                    df.to_csv(file_name, index=False, float_format=float_format)

                    # Limit the number of columns if this is end user report
                    """
                    if end_user_report:
                        cols = [col for col in df.columns if col in END_USER_COLS]
                        df = df[cols]
                    """

                    df_name_pretty = " ".join([word.capitalize() for word in df_name.split("_")])

                    # If it is not an end user report, or it is end user and it qualifies,
                    # we add the table to the html
                    if df_name in END_USER_METRIC_RESULT_KEYS or (not end_user_report):
                        html_str, makdown_str = self.publish_df_color_code_p_value(
                            df=df,
                            bg_cols=(
                                METRIC_NAME_COL,
                                "launch",
                                "delta_percent",
                                "p_value",
                                "delta_sum",
                            ),
                            df_name=df_name,
                            html_str=html_str,
                            markdown_str=markdown_str,
                            split_col="launch",
                        )

        print(f"\n*** `mea.publish`: html_file_name: {html_file_name}")
        print(f"\n*** `mea.publish`: markdown_file_name: {markdown_file_name}")
        if html_file_name is not None:
            with open(f"{write_path}/{html_file_name}", "w") as f:
                f.write(html_str)
                print(f"\n*** data was written to {write_path}/{html_file_name}.")

        if markdown_file_name is not None:
            with open(f"{write_path}/{markdown_file_name}", "w") as f:
                f.write(markdown_str)
                print(f"\n*** data was written to {write_path}/{markdown_file_name}.")

        return {
            "html_str": html_str,
            "markdown_str": markdown_str,
            "paths": paths,
            "file_names": file_names,
        }
